# upstream_checker/checker.py

import socket
import time
import http.client
from typing import Dict, List


def check_tcp(address: str, timeout: float, retries: int) -> bool:
    """
    Проверка доступности сервера по TCP.
    Ignores extra upstream options like 'max_fails' or 'fail_timeout'.
    """
    # Берем только host:port, игнорируем параметры
    host_port = address.split()[0]
    host, port = host_port.split(":")
    port = int(port)
    
    for _ in range(retries):
        try:
            with socket.create_connection((host, port), timeout=timeout):
                return True
        except (socket.timeout, ConnectionRefusedError, OSError):
            time.sleep(0.2)
    return False


def check_http(address: str, timeout: float, retries: int) -> bool:
    """
    Проверка доступности сервера по HTTP (GET /).
    Ignores extra upstream options like 'max_fails' or 'fail_timeout'.
    """
    host_port = address.split()[0]
    host, port = host_port.split(":")
    port = int(port)
    
    for _ in range(retries):
        try:
            conn = http.client.HTTPConnection(host, port, timeout=timeout)
            conn.request("GET", "/")
            resp = conn.getresponse()
            healthy = resp.status < 500
            conn.close()
            if healthy:
                return True
        except Exception:
            time.sleep(0.2)
            continue
    return False


def resolve_address(address: str) -> List[str]:
    """
    Резолвит адрес upstream сервера в IP-адреса.
    
    Args:
        address: Адрес в формате "host:port" или "host:port параметры"
        
    Returns:
        Список IP-адресов в формате "ip:port" или пустой список, если резолвинг не удался
    """
    try:
        host_port = address.split()[0]
        
        if ":" not in host_port:
            return []
            
        parts = host_port.rsplit(":", 1)
        if len(parts) != 2:
            return []
        host, port = parts
        
        # Если это уже IPv4 адрес, возвращаем как есть
        try:
            socket.inet_aton(host)
            return [host_port]
        except socket.error:
            pass
        
        # Проверяем IPv6 (в квадратных скобках)
        if host.startswith("[") and host.endswith("]"):
            ipv6_host = host[1:-1]
            try:
                socket.inet_pton(socket.AF_INET6, ipv6_host)
                return [host_port]
            except (socket.error, OSError):
                pass
        
        # Пытаемся резолвить DNS имя - получаем все IP-адреса
        try:
            # gethostbyname_ex возвращает (hostname, aliaslist, ipaddrlist)
            _, _, ipaddrlist = socket.gethostbyname_ex(host)
            # Фильтруем только IPv4 адреса (IPv6 обрабатываются отдельно)
            resolved_ips = []
            for ip in ipaddrlist:
                try:
                    # Проверяем, что это IPv4
                    socket.inet_aton(ip)
                    resolved_ips.append(f"{ip}:{port}")
                except socket.error:
                    pass
            return resolved_ips if resolved_ips else []
        except (socket.gaierror, OSError):
            return []
    except (ValueError, IndexError, AttributeError):
        return []


def resolve_upstreams(
    upstreams: Dict[str, List[str]]
) -> Dict[str, List[dict]]:
    """
    Резолвит DNS имена upstream-серверов в IP-адреса.
    
    Возвращает:
    {
        "backend": [
            {"address": "example.com:8080", "resolved": ["192.168.1.1:8080", "192.168.1.2:8080"]},
            {"address": "127.0.0.1:8080", "resolved": ["127.0.0.1:8080"]},
            {"address": "badhost:80", "resolved": []},
            ...
        ]
    }
    """
    results = {}
    for name, servers in upstreams.items():
        results[name] = []
        for srv in servers:
            resolved = resolve_address(srv)
            results[name].append({
                "address": srv,
                "resolved": resolved
            })
    return results


def check_upstreams(
    upstreams: Dict[str, List[str]],
    timeout: float = 2.0,
    retries: int = 1,
    mode: str = "tcp"
) -> Dict[str, List[dict]]:
    """
    Проверяет доступность upstream-серверов.
    mode: "tcp" (по умолчанию) или "http"
    
    Возвращает:
    {
        "backend": [
            {"address": "127.0.0.1:8080", "healthy": True},
            ...
        ]
    }
    """
    results = {}
    for name, servers in upstreams.items():
        results[name] = []
        for srv in servers:
            if mode.lower() == "http":
                healthy = check_http(srv, timeout, retries)
            else:
                healthy = check_tcp(srv, timeout, retries)
            results[name].append({"address": srv, "healthy": healthy})
    return results



import typer
from rich.console import Console
from rich.table import Table
from analyzer.conflicts import find_location_conflicts, find_listen_servername_conflicts
from analyzer.duplicates import find_duplicate_directives
from analyzer.empty_blocks import find_empty_blocks
from analyzer.warnings import find_warnings
from analyzer.unused import find_unused_variables
from parser.nginx_parser import parse_nginx_config
from analyzer.rewrite import find_rewrite_issues
from analyzer.dead_locations import find_dead_locations

app = typer.Typer()
console = Console()

# Карта советов и критичности для issue_type
ISSUE_META = {
    'location_conflict': ("Возможное пересечение location. Это не всегда ошибка: порядок и типы location могут быть корректны. Проверьте, что порядок и типы location соответствуют вашим ожиданиям. Если всё ок — игнорируйте предупреждение.", "medium"),
    'duplicate_directive': ("Оставьте только одну директиву с нужным значением в этом блоке.", "medium"),
    'empty_block': ("Удалите или заполните пустой блок.", "low"),
    'proxy_pass_no_scheme': ("Добавьте http:// или https:// в proxy_pass.", "medium"),
    'autoindex_on': ("Отключите autoindex, если не требуется публикация файлов.", "medium"),
    'if_block': ("Избегайте if внутри location, используйте map/try_files.", "medium"),
    'server_tokens_on': ("Отключите server_tokens для безопасности.", "low"),
    'ssl_missing': ("Укажите путь к SSL-сертификату/ключу.", "high"),
    'ssl_protocols_weak': ("Отключите устаревшие протоколы TLS.", "high"),
    'ssl_ciphers_weak': ("Используйте современные шифры.", "high"),
    'listen_443_no_ssl': ("Добавьте ssl к listen 443.", "high"),
    'listen_443_no_http2': ("Добавьте http2 к listen 443 для производительности.", "low"),
    'no_limit_req_conn': ("Добавьте limit_req/limit_conn для защиты от DDoS.", "medium"),
    'missing_security_header': ("Добавьте security-заголовок.", "medium"),
    'deprecated': ("Замените устаревшую директиву.", "medium"),
    'limit_too_small': ("Увеличьте лимит до рекомендуемого значения.", "medium"),
    'limit_too_large': ("Уменьшите лимит до разумного значения.", "medium"),
    'unused_variable': ("Удалите неиспользуемую переменную.", "low"),
    'listen_servername_conflict': ("Измените listen/server_name для устранения конфликта.", "high"),
    'rewrite_cycle': ("Проверьте rewrite на циклические правила.", "high"),
    'rewrite_conflict': ("Проверьте порядок и уникальность rewrite.", "medium"),
    'rewrite_no_flag': ("Добавьте last/break/redirect/permanent к rewrite.", "low"),
    'dead_location': ("Удалите неиспользуемый location или используйте его.", "low"),
}
SEVERITY_COLOR = {"high": "red", "medium": "orange3", "low": "yellow"}

def analyze(config_path: str = typer.Argument(..., help="Путь к nginx.conf")):
    """
    Анализирует конфигурацию Nginx на типовые проблемы и best practices.

    Показывает:
      - Конфликты location-ов
      - Дублирующиеся директивы
      - Пустые блоки
      - Потенциальные проблемы (proxy_pass без схемы, autoindex on, if, server_tokens on, SSL, лимиты, deprecated и др.)
      - Неиспользуемые переменные
      - Конфликты listen/server_name
      - Проблемы с rewrite
      - Мертвые location-ы

    Пример:
        nginx-lens analyze /etc/nginx/nginx.conf
    """
    try:
        tree = parse_nginx_config(config_path)
    except FileNotFoundError:
        console.print(f"[red]Файл {config_path} не найден. Проверьте путь к конфигу.[/red]")
        return
    except Exception as e:
        console.print(f"[red]Ошибка при разборе {config_path}: {e}[/red]")
        return
    conflicts = find_location_conflicts(tree)
    dups = find_duplicate_directives(tree)
    empties = find_empty_blocks(tree)
    warnings = find_warnings(tree)
    unused_vars = find_unused_variables(tree)
    listen_conflicts = find_listen_servername_conflicts(tree)
    rewrite_issues = find_rewrite_issues(tree)
    dead_locations = find_dead_locations(tree)

    table = Table(show_header=True, header_style="bold blue")
    table.add_column("issue_type")
    table.add_column("issue_description")
    table.add_column("solution")

    def add_row(issue_type, desc):
        solution, severity = ISSUE_META.get(issue_type, ("", "low"))
        color = SEVERITY_COLOR.get(severity, "yellow")
        table.add_row(f"[{color}]{issue_type}[/{color}]", desc, f"[{color}]{solution}[/{color}]")

    for c in conflicts:
        add_row("location_conflict", f"server: {c['server'].get('arg', '')} location: {c['location1']} ↔ {c['location2']}")
    for d in dups:
        loc = d.get('location')
        add_row("duplicate_directive", f"{d['directive']} ({d['args']}) — {d['count']} раз в блоке {d['block'].get('block', d['block'])}{' location: '+str(loc) if loc else ''}")
    for e in empties:
        add_row("empty_block", f"{e['block']} {e['arg'] or ''}")
    for w in warnings:
        t = w['type']
        if t == 'proxy_pass_no_scheme':
            add_row(t, f"proxy_pass без схемы: {w['value']}")
        elif t == 'autoindex_on':
            add_row(t, f"autoindex on в блоке {w['context'].get('block','')}")
        elif t == 'if_block':
            add_row(t, f"директива if внутри блока {w['context'].get('block','')}")
        elif t == 'server_tokens_on':
            add_row(t, f"server_tokens on в блоке {w['context'].get('block','')}")
        elif t == 'ssl_missing':
            add_row(t, f"{w['directive']} не указан")
        elif t == 'ssl_protocols_weak':
            add_row(t, f"ssl_protocols содержит устаревшие протоколы: {w['value']}")
        elif t == 'ssl_ciphers_weak':
            add_row(t, f"ssl_ciphers содержит слабые шифры: {w['value']}")
        elif t == 'listen_443_no_ssl':
            add_row(t, f"listen без ssl: {w['value']}")
        elif t == 'listen_443_no_http2':
            add_row(t, f"listen 443 без http2: {w['value']}")
        elif t == 'no_limit_req_conn':
            add_row(t, f"server без limit_req/limit_conn")
        elif t == 'missing_security_header':
            add_row(t, f"отсутствует security header: {w['value']}")
        elif t == 'deprecated':
            add_row(t, f"устаревшая директива: {w['directive']} — {w['value']}")
        elif t == 'limit_too_small':
            add_row(t, f"слишком маленькое значение: {w['directive']} = {w['value']}")
        elif t == 'limit_too_large':
            add_row(t, f"слишком большое значение: {w['directive']} = {w['value']}")
    for v in unused_vars:
        add_row("unused_variable", v['name'])
    for c in listen_conflicts:
        add_row("listen_servername_conflict", f"server1: {c['server1'].get('arg','')} server2: {c['server2'].get('arg','')} listen: {','.join(c['listen'])} server_name: {','.join(c['server_name'])}")
    for r in rewrite_issues:
        add_row(r['type'], r['value'])
    for l in dead_locations:
        add_row("dead_location", f"server: {l['server'].get('arg','')} location: {l['location'].get('arg','')}")

    if table.row_count == 0:
        console.print("[green]Проблем не найдено[/green]")
    else:
        console.print(table) 


from typing import List, Dict, Any

def diff_trees(tree1, tree2) -> List[Dict[str, Any]]:
    """
    Сравнивает два дерева директив. Возвращает список отличий.
    Формат: [{type, path, value1, value2}]
    type: 'added', 'removed', 'changed'
    path: путь до директивы (например, ['http', 'server', 'location /api'])
    """
    diffs = []
    _diff_blocks(tree1.directives, tree2.directives, [], diffs)
    return diffs

def _diff_blocks(d1, d2, path, diffs):
    # Индексируем по (block,arg) или (directive,args) или (upstream)
    def key(d):
        if 'block' in d:
            return ('block', d['block'], d.get('arg'))
        if 'directive' in d:
            return ('directive', d['directive'], d.get('args'))
        if 'upstream' in d:
            return ('upstream', d['upstream'])
        return ('other', str(d))
    map1 = {key(x): x for x in d1}
    map2 = {key(x): x for x in d2}
    all_keys = set(map1) | set(map2)
    for k in all_keys:
        v1 = map1.get(k)
        v2 = map2.get(k)
        p = path + [k[1] if k[0] != 'directive' else f"{k[1]} {k[2] or ''}".strip()]
        if v1 and not v2:
            diffs.append({'type': 'removed', 'path': p, 'value1': v1, 'value2': None})
        elif v2 and not v1:
            diffs.append({'type': 'added', 'path': p, 'value1': None, 'value2': v2})
        else:
            # Рекурсивно сравниваем блоки
            if 'directives' in v1 and 'directives' in v2:
                _diff_blocks(v1['directives'], v2['directives'], p, diffs)
            elif v1 != v2:
                diffs.append({'type': 'changed', 'path': p, 'value1': v1, 'value2': v2}) 


import typer
from rich.console import Console
from rich.table import Table
from analyzer.diff import diff_trees
from parser.nginx_parser import parse_nginx_config
import difflib

app = typer.Typer()
console = Console()

def diff(
    config1: str = typer.Argument(..., help="Первый nginx.conf"),
    config2: str = typer.Argument(..., help="Второй nginx.conf")
):
    """
    Сравнивает две конфигурации Nginx и выводит отличия построчно side-by-side с подсветкой.

    Пример:
        nginx-lens diff /etc/nginx/nginx.conf /etc/nginx/nginx.conf.bak
    """
    try:
        with open(config1) as f1, open(config2) as f2:
            lines1 = f1.readlines()
            lines2 = f2.readlines()
    except FileNotFoundError as e:
        console.print(f"[red]Файл {e.filename} не найден. Проверьте путь к конфигу.[/red]")
        return
    except Exception as e:
        console.print(f"[red]Ошибка при чтении файлов: {e}[/red]")
        return
    maxlen = max(len(lines1), len(lines2))
    # Выравниваем длины
    lines1 += [''] * (maxlen - len(lines1))
    lines2 += [''] * (maxlen - len(lines2))
    table = Table(show_header=True, header_style="bold blue")
    table.add_column("№", style="dim", width=4)
    table.add_column("Config 1", style="white")
    table.add_column("№", style="dim", width=4)
    table.add_column("Config 2", style="white")
    for i in range(maxlen):
        l1 = lines1[i].rstrip('\n')
        l2 = lines2[i].rstrip('\n')
        n1 = str(i+1) if l1 else ''
        n2 = str(i+1) if l2 else ''
        if l1 == l2:
            table.add_row(n1, l1, n2, l2)
        else:
            style1 = "red" if l1 else "on red"
            style2 = "green" if l2 else "on green"
            table.add_row(f"[bold]{n1}[/bold]" if l1 else n1, f"[{style1}]{l1}[/{style1}]" if l1 or l2 else '', f"[bold]{n2}[/bold]" if l2 else n2, f"[{style2}]{l2}[/{style2}]" if l1 or l2 else '')
    console.print(table) 


from analyzer.base import Analyzer
from typing import List, Dict, Any

def find_empty_blocks(tree) -> List[Dict[str, Any]]:
    """
    Находит пустые блоки (без вложенных директив).
    Возвращает список: [{block, arg}]
    """
    analyzer = Analyzer(tree)
    empties = []
    for d, parent in analyzer.walk():
        if 'block' in d and (not d.get('directives')):
            empties.append({'block': d.get('block'), 'arg': d.get('arg')})
    return empties 


import tempfile
import os
from parser.nginx_parser import parse_nginx_config

def test_simple_upstream():
    conf = """
    upstream backend {
        server 127.0.0.1:8080;
        server 10.0.0.1:80;
    }
    """
    with tempfile.NamedTemporaryFile("w+", delete=False) as f:
        f.write(conf)
        f.flush()
        tree = parse_nginx_config(f.name)
    ups = tree.get_upstreams()
    assert "backend" in ups
    assert set(ups["backend"]) == {"127.0.0.1:8080", "10.0.0.1:80"}
    os.unlink(f.name)

def test_upstream_with_include():
    conf_main = """
    include sub.conf;
    """
    conf_sub = """
    upstream api {
        server api1:9000;
    }
    """
    with tempfile.TemporaryDirectory() as d:
        main_path = os.path.join(d, "nginx.conf")
        sub_path = os.path.join(d, "sub.conf")
        with open(main_path, "w") as f:
            f.write(conf_main)
        with open(sub_path, "w") as f:
            f.write(conf_sub)
        tree = parse_nginx_config(main_path)
        ups = tree.get_upstreams()
        assert "api" in ups
        assert ups["api"] == ["api1:9000"]

def test_nested_blocks_and_comments():
    conf = """
    # http block
    http {
        upstream u1 { server 1.1.1.1:80; } # inline
        server { # comment
            listen 80;
            location /api {
                proxy_pass http://u1;
            }
        }
    }
    """
    with tempfile.NamedTemporaryFile("w+", delete=False) as f:
        f.write(conf)
        f.flush()
        tree = parse_nginx_config(f.name)
    ups = tree.get_upstreams()
    assert "u1" in ups
    assert ups["u1"] == ["1.1.1.1:80"]
    os.unlink(f.name) 


import typer
from rich.console import Console
from rich.table import Table
from upstream_checker.checker import check_upstreams, resolve_upstreams
from parser.nginx_parser import parse_nginx_config

app = typer.Typer()
console = Console()

def health(
    config_path: str = typer.Argument(..., help="Путь к nginx.conf"),
    timeout: float = typer.Option(2.0, help="Таймаут проверки (сек)"),
    retries: int = typer.Option(1, help="Количество попыток"),
    mode: str = typer.Option("tcp", help="Режим проверки: tcp или http", case_sensitive=False),
    resolve: bool = typer.Option(False, "--resolve", "-r", help="Показать резолвленные IP-адреса"),
):
    """
    Проверяет доступность upstream-серверов, определённых в nginx.conf. Выводит таблицу.

    Пример:
        nginx-lens health /etc/nginx/nginx.conf
        nginx-lens health /etc/nginx/nginx.conf --timeout 5 --retries 3 --mode http
        nginx-lens health /etc/nginx/nginx.conf --resolve
    """
    try:
        tree = parse_nginx_config(config_path)
    except FileNotFoundError:
        console.print(f"[red]Файл {config_path} не найден. Проверьте путь к конфигу.[/red]")
        return
    except Exception as e:
        console.print(f"[red]Ошибка при разборе {config_path}: {e}[/red]")
        return

    upstreams = tree.get_upstreams()
    results = check_upstreams(upstreams, timeout=timeout, retries=retries, mode=mode.lower())
    
    # Если нужно показать резолвленные IP-адреса
    resolved_info = {}
    if resolve:
        resolved_info = resolve_upstreams(upstreams)

    table = Table(show_header=True, header_style="bold blue")
    table.add_column("Address")
    table.add_column("Status")
    if resolve:
        table.add_column("Resolved IP")

    for name, servers in results.items():
        for srv in servers:
            status = "Healthy" if srv["healthy"] else "Unhealthy"
            color = "green" if srv["healthy"] else "red"
            
            if resolve:
                resolved_list = []
                if name in resolved_info:
                    for resolved_srv in resolved_info[name]:
                        if resolved_srv["address"] == srv["address"]:
                            resolved_list = resolved_srv["resolved"]
                            break
                
                if resolved_list:
                    # Показываем все IP-адреса через запятую
                    resolved_str = ", ".join(resolved_list)
                    table.add_row(srv["address"], f"[{color}]{status}[/{color}]", f"[green]{resolved_str}[/green]")
                else:
                    table.add_row(srv["address"], f"[{color}]{status}[/{color}]", "[yellow]Failed to resolve[/yellow]")
            else:
                table.add_row(srv["address"], f"[{color}]{status}[/{color}]")

    console.print(table)



def tree_to_markdown(directives, level=0):
    lines = []
    prefix = '  ' * level + '- '
    for d in directives:
        if 'block' in d:
            label = f"{d['block']} {d.get('arg') or ''}".strip()
            lines.append(f"{prefix}{label}")
            if d.get('directives'):
                lines.append(tree_to_markdown(d['directives'], level+1))
        elif 'upstream' in d:
            label = f"upstream {d['upstream']}"
            lines.append(f"{prefix}{label}")
            for srv in d.get('servers', []):
                lines.append(f"{'  '*(level+1)}- server {srv}")
        elif 'directive' in d:
            lines.append(f"{prefix}{d['directive']} {d.get('args','')}")
    return '\n'.join(lines) 


from analyzer.base import Analyzer
from typing import List, Dict, Any

def find_duplicate_directives(tree) -> List[Dict[str, Any]]:
    """
    Находит дублирующиеся директивы внутри одного блока (без вложенности).
    Возвращает список: [{block, directive, count, location}]
    """
    analyzer = Analyzer(tree)
    duplicates = []
    for d, parent in analyzer.walk():
        if 'directives' in d:
            # Считаем только прямые дочерние директивы (без вложенных блоков)
            seen = {}
            for sub in d['directives']:
                if 'directive' in sub:
                    key = (sub['directive'], str(sub.get('args')))
                    seen[key] = seen.get(key, 0) + 1
            for (directive, args), count in seen.items():
                if count > 1:
                    duplicates.append({
                        'block': d,
                        'directive': directive,
                        'args': args,
                        'count': count,
                        'location': d.get('arg')
                    })
    return duplicates 